@@ -27,9 +27,9 @@ from difflib import SequenceMatcher
 from pickle import dumps
 
 
-__all__ = ['Differ', 'Patcher', 'diff', 'patch']
+__all__ = ['Differ', 'Iterator', 'Patcher', 'diff', 'patch']
 
-__version__ = '0.4'
+__version__ = '0.5'
 __author__ = 'Michael Samoglyadov'
 __license__ = 'Apache License, Version 2.0'
 __website__ = 'https://github.com/mr-mixas/Nested-Diff.py'
@@ -42,6 +42,7 @@ class Differ(object):
     Resulting diff is a dict and may contain following keys:
     `A` stands for 'added', it's value - added item.
     `D` means 'different' and contains subdiff.
+    `E` diffed entity (optional), value - empty instance of entity's class.
     `I` index for sequence item, used only when prior item was omitted.
     `N` is a new value for changed item.
     `O` is a changed item's old value.
@@ -49,10 +50,12 @@ class Differ(object):
     `U` represent unchanged item.
 
     Diff metadata alternates with actual data; simple types specified as is,
-    dicts, lists, sets and tuples contain subdiffs for their items with native
-    for such types addressing: indexes for lists and tuples and keys for
-    dictionaries. Each status type, except `D` and `I`, may be omitted during
-    diff computation.
+    dicts, lists and tuples contain subdiffs for their items with native for
+    such types addressing: indexes for lists and tuples and keys for
+    dictionaries. Each status type, except `D`. `E` and `I`, may be omitted
+    during diff computation. `E` tag is used with `D` when entity unable to
+    contain diff by itself (set, frozenset); `D` contain a list of subdiffs
+    in this case.
 
     Example:
 
@@ -85,7 +88,7 @@ class Differ(object):
 
     """
     def __init__(self, A=True, N=True, O=True, R=True, U=True, trimR=False,
-                 diff_method=None):
+                 diff_method=None, multiline_diff_context=-1):
         """
         Construct Differ.
 
@@ -101,6 +104,9 @@ class Differ(object):
         Differ opts will be passed as arguments, diff expected for output.
         Disabled (`None`) by default.
 
+        `multiline_diff_context` defines amount of context lines for multiline
+        string diffs, multiline diffs disabled when value is negative.
+
         """
         self.__diff_method = diff_method
         self.lcs = SequenceMatcher(isjunk=None, autojunk=False)
@@ -112,12 +118,25 @@ class Differ(object):
         self.op_u = U
         self.op_trim_r = trimR
 
+        self.__differs = {
+            dict: self.diff_dict,
+            frozenset: self.diff_set,
+            list: self.diff_list,
+            set: self.diff_set,
+            tuple: self.diff_tuple,
+        }
+
+        if multiline_diff_context >= 0:
+            self.__differs[str] = self.diff_multiline
+            self.multiline_diff_context = multiline_diff_context
+
     def diff(self, a, b):
         """
         Compute diff for two arbitrary objects.
 
-        This method is a dispatcher and calls `diff_dicts` for dicts,
-        `diff_lists` for lists and so forth.
+        This method is a dispatcher and calls registered diff method for each
+        diffed values pair according to their type. `diff__default` called for
+        non-registered types. Args and kwargs passed to called method as is.
 
         :param a: First object to diff.
         :param b: Second object to diff.
@@ -135,23 +154,11 @@ class Differ(object):
                 diff_method=self.__diff_method,
             )
 
-        if a == b:
-            return {'U': a} if self.op_u else {}
-
-        if isinstance(a, dict) and isinstance(a, type(b)):
-            return self.diff_dicts(a, b)
-
-        if isinstance(a, list) and isinstance(a, type(b)):
-            return self.diff_lists(a, b)
-
-        if isinstance(a, set) and isinstance(a, type(b)):
-            return self.diff_sets(a, b)
-
-        if isinstance(a, tuple) and isinstance(a, type(b)):
-            return self.diff_tuples(a, b)
-
-        if isinstance(a, frozenset) and isinstance(a, type(b)):
-            return self.diff_frozensets(a, b)
+        if a.__class__ is b.__class__:
+            if a == b:
+                return {'U': a} if self.op_u else {}
+            else:
+                return self.get_differ(a.__class__)(a, b)
 
         return self.diff__default(a, b)
 
@@ -169,7 +176,7 @@ class Differ(object):
 
         return ret
 
-    def diff_dicts(self, a, b):
+    def diff_dict(self, a, b):
         """
         Compute diff for two dicts.
 
@@ -184,32 +191,32 @@ class Differ(object):
         >>>
 
         """
-        ret = {'D': {}}
+        dif = {}
 
         for key in set(list(a) + list(b)):
             if key in a and key in b:
                 if a[key] == b[key]:
                     if self.op_u:
-                        ret['D'][key] = {'U': a[key]}
+                        dif[key] = {'U': a[key]}
                 else:
                     subdiff = self.diff(a[key], b[key])
                     if subdiff:
-                        ret['D'][key] = subdiff
+                        dif[key] = subdiff
 
             elif key in a:  # removed
                 if self.op_r:
-                    ret['D'][key] = {'R': None if self.op_trim_r else a[key]}
+                    dif[key] = {'R': None if self.op_trim_r else a[key]}
 
             else:  # added
                 if self.op_a:
-                    ret['D'][key] = {'A': b[key]}
+                    dif[key] = {'A': b[key]}
 
-        if not ret['D']:
-            del ret['D']
+        if dif:
+            return {'D': dif}
 
-        return ret
+        return dif
 
-    def diff_lists(self, a, b):
+    def diff_list(self, a, b):
         """
         Compute diff for two lists.
 
@@ -224,10 +231,10 @@ class Differ(object):
         >>>
 
         """
-        self.lcs.set_seq1([dumps(i, -1) for i in a])
-        self.lcs.set_seq2([dumps(i, -1) for i in b])
+        self.lcs.set_seq1(tuple(dumps(i, -1) for i in a))
+        self.lcs.set_seq2(tuple(dumps(i, -1) for i in b))
 
-        ret = {'D': []}
+        dif = []
         i = j = 0
         force_index = False
 
@@ -235,9 +242,9 @@ class Differ(object):
             while i < ai and j < bj:
                 subdiff = self.diff(a[i], b[j])
                 if subdiff:
-                    ret['D'].append(subdiff)
+                    dif.append(subdiff)
                     if force_index:
-                        ret['D'][-1]['I'] = i
+                        dif[-1]['I'] = i
                         force_index = False
                 else:
                     force_index = True
@@ -247,9 +254,9 @@ class Differ(object):
 
             while i < ai:  # removed
                 if self.op_r:
-                    ret['D'].append({'R': None if self.op_trim_r else a[i]})
+                    dif.append({'R': None if self.op_trim_r else a[i]})
                     if force_index:
-                        ret['D'][-1]['I'] = i
+                        dif[-1]['I'] = i
                         force_index = False
                 else:
                     force_index = True
@@ -258,23 +265,66 @@ class Differ(object):
 
             while j < bj:  # added
                 if self.op_a:
-                    ret['D'].append({'A': b[j]})
+                    dif.append({'A': b[j]})
                     if force_index:
-                        ret['D'][-1]['I'] = i
+                        dif[-1]['I'] = i
                         force_index = False
                 else:
                     force_index = True
 
                 j += 1
 
-        if not ret['D']:
-            del ret['D']
+        if dif:
+            return {'D': dif}
 
-        return ret
+        return {}
 
-    def diff_sets(self, a, b):
+    def diff_multiline(self, a, b):
         """
-        Compute diff for two sets.
+        Compute diff for multiline strings.
+        Highly experimental! Format may be changed at any time.
+
+        """
+        lines_a = a.split('\n', -1)
+        lines_b = b.split('\n', -1)
+
+        if len(lines_a) == len(lines_b) == 1:
+            return self.diff__default(a, b)
+
+        dif = []
+        self.lcs.set_seq1(lines_a)
+        self.lcs.set_seq2(lines_b)
+
+        for group in self.lcs.get_grouped_opcodes(self.multiline_diff_context):
+            dif.append({
+                'I': [
+                    group[0][1], group[-1][2],
+                    group[0][3], group[-1][4],
+                ]
+            })
+
+            for op, i1, i2, j1, j2 in group:
+                if op == 'equal':
+                    for line in lines_a[i1:i2]:
+                        dif.append({'U': line})
+                    continue
+
+                if op in {'replace', 'delete'}:
+                    for line in lines_a[i1:i2]:
+                        dif.append({'R': line})
+
+                if op in {'replace', 'insert'}:
+                    for line in lines_b[j1:j2]:
+                        dif.append({'A': line})
+
+        if dif:
+            return {'D': dif, 'E': a.__class__()}
+
+        return {}
+
+    def diff_set(self, a, b):
+        """
+        Compute diff for two [frozen]sets.
 
         :param a: First set to diff.
         :param b: Second set to diff.
@@ -283,53 +333,31 @@ class Differ(object):
         >>> b = {2, 3}
         >>>
         >>> Differ(U=False).diff_sets(a, b)
-        {'D': {{'R': 1}, {'A': 3}}}
+        {'D': [{'R': 1}, {'A': 3}], 'E': set()}
         >>>
 
         """
-        ret = {'D': set()}
+        dif = []
 
         for i in a.union(b):
             if i in a and i in b:
                 if self.op_u:
-                    ret['D'].add(_hdict('U', i))
+                    dif.append({'U': i})
 
             elif i in a:  # removed
                 if self.op_r:
-                    ret['D'].add(_hdict('R', None if self.op_trim_r else i))
+                    dif.append({'R': None if self.op_trim_r else i})
 
             else:  # added
                 if self.op_a:
-                    ret['D'].add(_hdict('A', i))
+                    dif.append({'A': i})
 
-        if not ret['D']:
-            del ret['D']
+        if dif:
+            return {'D': dif, 'E': a.__class__()}
 
-        return ret
+        return {}
 
-    def diff_frozensets(self, a, b):
-        """
-        Compute diff for two frozen sets.
-
-        :param a: First frozenset to diff.
-        :param b: Second frozenset to diff.
-
-        >>> a = frozenset((1, 2))
-        >>> b = frozenset((2, 3))
-        >>>
-        >>> Differ(U=False).diff_frozensets(a, b)
-        {'D': frozenset({{'R': 1}, {'A': 3}})}
-        >>>
-
-        """
-        ret = self.diff_sets(a, b)
-
-        if 'D' in ret:
-            ret['D'] = frozenset(ret['D'])
-
-        return ret
-
-    def diff_tuples(self, a, b):
+    def diff_tuple(self, a, b):
         """
         Compute diff for two tuples.
 
@@ -344,25 +372,34 @@ class Differ(object):
         >>>
 
         """
-        ret = self.diff_lists(a, b)
+        ret = self.diff_list(a, b)
 
         if 'D' in ret:
             ret['D'] = tuple(ret['D'])
 
         return ret
 
+    def get_differ(self, cls):
+        """
+        Return diff method for specified type.
 
-class _hdict(dict):
-    """
-    Hashable dict, for internal use only.
-    """
-    def __init__(self, op, val):
-        dict.__init__(self)
-        self[op] = val
-        self.__hash = hash((op, val))
+        :param cls: diffed object type.
 
-    def __hash__(self):
-        return self.__hash
+        """
+        try:
+            return self.__differs[cls]
+        except KeyError:
+            return self.diff__default
+
+    def set_differ(self, cls, method):
+        """
+        Set differ for specified data type.
+
+        :param cls: diffed object type.
+        :param method: diff method.
+
+        """
+        self.__differs[cls] = method
 
 
 class Patcher(object):
@@ -382,12 +419,33 @@ class Patcher(object):
         """
         self.__patch_method = patch_method
 
+        self.__patchers = {
+            dict: self.patch_dict,
+            frozenset: self.patch_frozenset,
+            list: self.patch_list,
+            set: self.patch_set,
+            tuple: self.patch_tuple,
+        }
+
+    def get_patcher(self, cls):
+        """
+        Return patch method for specified type.
+
+        :param cls: patched object type.
+
+        """
+        try:
+            return self.__patchers[cls]
+        except KeyError:
+            raise NotImplementedError("unsupported object type") from None
+
     def patch(self, target, ndiff):
         """
         Return patched object.
 
-        This method is a dispatcher and calls `patch_dict` for dicts,
-        `patch_list` for lists and so forth.
+        This method is a dispatcher and calls registered patch method for each
+        patched value according to it's type. `patch__default` called for
+        non-registered types. Args and kwargs passed to called method as is.
 
         :param target: Object to patch.
         :param ndiff: Nested diff.
@@ -398,36 +456,14 @@ class Patcher(object):
             return getattr(target, self.__patch_method)(ndiff)
 
         if 'D' in ndiff:
-            if isinstance(ndiff['D'], dict):
-                return self.patch_dict(target, ndiff)
-
-            if isinstance(ndiff['D'], list):
-                return self.patch_list(target, ndiff)
-
-            if isinstance(ndiff['D'], set):
-                return self.patch_set(target, ndiff)
-
-            if isinstance(ndiff['D'], tuple):
-                return self.patch_tuple(target, ndiff)
-
-            if isinstance(ndiff['D'], frozenset):
-                return self.patch_frozenset(target, ndiff)
-
-            return self.patch__default(target, ndiff)
-
+            return self.get_patcher(
+                ndiff.get('E', ndiff['D']).__class__
+            )(target, ndiff)
         elif 'N' in ndiff:
             return ndiff['N']
-
         else:
             return target
 
-    def patch__default(self, target, ndiff):
-        """
-        Patch containers without dedicated methods.
-
-        """
-        raise NotImplementedError("unsupported object type")
-
     def patch_dict(self, target, ndiff):
         """
         Return patched dict.
@@ -474,7 +510,8 @@ class Patcher(object):
 
         return target
 
-    def patch_set(self, target, ndiff):
+    @staticmethod
+    def patch_set(target, ndiff):
         """
         Return patched set.
 
@@ -510,6 +547,118 @@ class Patcher(object):
         """
         return frozenset(self.patch_set(set(target), ndiff))
 
+    def set_patcher(self, cls, method):
+        """
+        Set patcher for specified data type.
+
+        :param cls: patched object type.
+        :param method: patch method.
+
+        """
+        self.__patchers[cls] = method
+
+
+class Iterator(object):
+    """
+    Nested diff iterator.
+
+    """
+    def __init__(self, sort_keys=False):
+        """
+        Construct iterator.
+
+        If `sort_keys` is `True`, then the output for mappings will be
+        sorted by key. Disabled by default.
+
+        """
+        self.sort_keys = sort_keys
+
+        self.__iters = {
+            dict: self._iter_mapping,
+            list: self._iter_sequence,
+            tuple: self._iter_sequence,
+        }
+
+    def _iter_mapping(self, value):
+        """
+        Iterate over dict-like objects.
+
+        :param value: mapping.
+
+        """
+        items = sorted(value.items()) if self.sort_keys else value.items()
+        for key, val in items:
+            yield value.__class__, key, val
+
+    @staticmethod
+    def _iter_sequence(value):
+        """
+        Iterate over lists, tuples and other sequences.
+
+        :param value: sequence.
+
+        """
+        idx = 0
+        for item in value:
+            if 'I' in item:
+                idx = item['I']
+
+            yield value.__class__, idx, item
+
+            idx += 1
+
+    def get_iter(self, value):
+        """
+        Return apropriate iterator for passed diff value.
+
+        """
+        try:
+            return self.__iters[value.__class__](value)
+        except KeyError:
+            raise NotImplementedError from None
+
+    def set_iter(self, type_, method):
+        """
+        Set generator for specified data type.
+
+        :param type_: data type.
+        :param method: method.
+
+        Generator should yield tuples with three items: container_type, pointer
+        and subdiff.
+
+        """
+        self.__iters[type_] = method
+
+    def iterate(self, ndiff, depth=0):
+        """
+        Return tuples with depth, container_type, pointer and subdiff for each
+        nested diff.
+
+        :param ndiff: Nested diff.
+
+        """
+        stack = [((None, None, _) for _ in (ndiff,))]
+
+        while True:
+            try:
+                container_type, pointer, subdiff = next(stack[-1])
+            except StopIteration:
+                stack.pop()
+
+                if stack:
+                    depth -= 1
+                    continue
+                else:
+                    break
+
+            yield depth, container_type, pointer, subdiff
+
+            if 'D' in subdiff:
+                if 'E' not in subdiff:
+                    stack.append(self.get_iter(subdiff['D']))
+                    depth += 1
+
 
 def diff(a, b, **kwargs):
     """
